@@ -26,7 +26,7 @@
 
 extern const char *minor_str[];
 
-#define VERSION "v5.3"
+#define VERSION "v5.4"
 
 #undef DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
@@ -43,16 +43,14 @@ extern const char *minor_str[];
 #endif
 #endif
 
-#define TIMING_MB		64
-#define TIMING_BUF_MB		1
+#define TIMING_BUF_MB		2
 #define TIMING_BUF_BYTES	(TIMING_BUF_MB * 1024 * 1024)
-#define TIMING_BUF_COUNT	(timing_MB / TIMING_BUF_MB)
 #define BUFCACHE_FACTOR		2
 
 char *progname;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
 static int flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_xt_hd = 0;
-static int do_ctimings, do_timings = 0, timing_MB = TIMING_MB;
+static int do_ctimings, do_timings = 0;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;
@@ -160,6 +158,7 @@ static void dump_identity (const struct hd_driveid *id)
 	int i;
 	char pmodes[64] = {0,}, dmodes[128]={0,}, umodes[128]={0,};
 	const unsigned short int *id_regs= (const void*) id;
+	unsigned long capacity;
 
 	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s",
 		id->model, id->fw_rev, id->serial_no);
@@ -192,16 +191,11 @@ static void dump_identity (const struct hd_driveid *id)
 	if (!(id->field_valid&1))
 		printf(" (maybe):");
 #if __BYTE_ORDER == __BIG_ENDIAN
-	printf(" CurCHS=%u/%u/%u, CurSects=%u",
-		id->cur_cyls, id->cur_heads, id->cur_sectors,
-		(id->cur_capacity0 << 16) | id->cur_capacity1
-		);
+	capacity = (id->cur_capacity0 << 16) | id->cur_capacity1;
 #else
-	printf(" CurCHS=%u/%u/%u, CurSects=%u",
-		id->cur_cyls, id->cur_heads, id->cur_sectors,
-		(id->cur_capacity1 << 16) | id->cur_capacity0
-		);
+	capacity = (id->cur_capacity1 << 16) | id->cur_capacity0;
 #endif
+	printf(" CurCHS=%u/%u/%u, CurSects=%lu", id->cur_cyls, id->cur_heads, id->cur_sectors, capacity);
 	printf(", LBA=%s", YN(id->capability&2));
 	if (id->capability&2)
  		printf(", LBAsects=%u", id->lba_capacity);
@@ -283,19 +277,22 @@ static void dump_identity (const struct hd_driveid *id)
 	if (id_regs[82]&0x20)
 		printf(" WriteCache=%s",(id_regs[85]&0x20) ? "enabled" : "disabled");
 #ifdef __NEW_HD_DRIVE_ID
-	if ((id->minor_rev_num) || (id->major_rev_num)) {
+	if ((id->minor_rev_num && id->minor_rev_num <= 31) || (id->major_rev_num && id->minor_rev_num <= 31)) {
 		printf("\n Drive conforms to: ");
 		if (id->minor_rev_num <= 31)
 			printf("%s: ", minor_str[id->minor_rev_num]);
 		else
 			printf("unknown: ");
-		for (i=0; i<=15; i++) {
-			if (id->major_rev_num & (1<<i))
-				printf(" %u", i);
+		if (id->major_rev_num < 31) {
+			for (i=0; i <= 15; i++) {
+				if (id->major_rev_num & (1<<i))
+					printf(" %u", i);
+			}
 		}
 	}
 #endif /* __NEW_HD_DRIVE_ID */
 	printf("\n");	
+	printf("\n * signifies the current active mode\n");
 	printf("\n");
 }
 
@@ -343,14 +340,13 @@ int read_big_block (int fd, char *buf)
 	return 0;
 }
 
-static double correction = 0.0;
-
 void time_cache (int fd)
 {
-	int i;
 	char *buf;
 	struct itimerval e1, e2;
 	int shmid;
+	double elapsed, elapsed2;
+	unsigned int iterations, total_MB;
 
 	if ((shmid = shmget(IPC_PRIVATE, TIMING_BUF_BYTES, 0600)) == -1) {
 		perror ("could not allocate sharedmem buf");
@@ -373,12 +369,7 @@ void time_cache (int fd)
 	sync();
 	sleep(3);
 
-	/* Calculate a correction factor for the basic
-	 * overhead of doing a read() from the buffer cache.
-	 * To do this, we read the data once to "cache it" and
-	 * to force full preallocation of our timing buffer,
-	 * and then we re-read it 10 times while timing it.
-	 *
+	/*
 	 * getitimer() is used rather than gettimeofday() because
 	 * it is much more consistent (on my machine, at least).
 	 */
@@ -392,34 +383,42 @@ void time_cache (int fd)
 	sync();
 	sleep(1);
 
-	/* Time re-reading from the buffer-cache */
+	/* Now do the timing */
+	iterations = 0;
 	getitimer(ITIMER_REAL, &e1);
-	for (i = (BUFCACHE_FACTOR * TIMING_BUF_COUNT) ; i > 0; --i) {
-		if (seek_to_zero (fd)) goto quit;
-		if (read_big_block (fd, buf)) goto quit;
-	}
-	getitimer(ITIMER_REAL, &e2);
-	correction = (e1.it_value.tv_sec - e2.it_value.tv_sec)
+	do {
+		++iterations;
+		if (seek_to_zero (fd) || read_big_block (fd, buf))
+			goto quit;
+		getitimer(ITIMER_REAL, &e2);
+		elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
+		 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
+	} while (elapsed < 2.0);
+	total_MB = iterations * TIMING_BUF_MB;
+
+	elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
 	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
 
-	/* Now remove the lseek() from the correction factor */
+	/* Now remove the lseek() and getitimer() overheads from the elapsed time */
 	getitimer(ITIMER_REAL, &e1);
-	for (i = (BUFCACHE_FACTOR * TIMING_BUF_COUNT) ; i > 0; --i) {
-		if (seek_to_zero (fd)) goto quit;
-	}
-	getitimer(ITIMER_REAL, &e2);
-	correction -= (e1.it_value.tv_sec - e2.it_value.tv_sec)
-	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
+	do {
+		if (seek_to_zero (fd))
+			goto quit;
+		getitimer(ITIMER_REAL, &e2);
+		elapsed2 = (e1.it_value.tv_sec - e2.it_value.tv_sec)
+		 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
+	} while (--iterations);
 
-	if ((BUFCACHE_FACTOR * timing_MB) >= correction)  /* more than 1MB/s */
-		printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
-			(BUFCACHE_FACTOR * timing_MB), correction,
-			(BUFCACHE_FACTOR * timing_MB) / correction);
+	elapsed -= elapsed2;
+
+	if ((BUFCACHE_FACTOR * total_MB) >= elapsed)  /* more than 1MB/s */
+		printf("%3u MB in %5.2f seconds = %6.2f MB/sec\n",
+			(BUFCACHE_FACTOR * total_MB), elapsed,
+			(BUFCACHE_FACTOR * total_MB) / elapsed);
 	else
-		printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
-			(BUFCACHE_FACTOR * timing_MB), correction,
-			(BUFCACHE_FACTOR * timing_MB) / correction * 1024);
-	correction /= BUFCACHE_FACTOR;
+		printf("%3u MB in %5.2f seconds = %6.2f kB/sec\n",
+			(BUFCACHE_FACTOR * total_MB), elapsed,
+			(BUFCACHE_FACTOR * total_MB) / elapsed * 1024);
 
 	flush_buffer_cache(fd);
 	sleep(1);
@@ -430,11 +429,22 @@ quit:
 
 void time_device (int fd)
 {
-	int i;
 	char *buf;
 	double elapsed;
 	struct itimerval e1, e2;
 	int shmid;
+	unsigned int max_iterations = 1024, total_MB, iterations;
+	static long parm;
+
+	//
+	// get device size
+	//
+	if (do_ctimings || do_timings) {
+		if (ioctl(fd, BLKGETSIZE, &parm))
+			perror(" BLKGETSIZE failed");
+		else
+			max_iterations = parm / (2 * 1024) / TIMING_BUF_MB;
+	}
 
 	if ((shmid = shmget(IPC_PRIVATE, TIMING_BUF_BYTES, 0600)) == -1) {
 		perror ("could not allocate sharedmem buf");
@@ -467,35 +477,24 @@ void time_device (int fd)
 	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
 
 	/* Now do the timings for real */
+	iterations = 0;
 	getitimer(ITIMER_REAL, &e1);
-	for (i = TIMING_BUF_COUNT; i > 0; --i) {
-		if (read_big_block (fd, buf)) goto quit;
-	}
-	getitimer(ITIMER_REAL, &e2);
+	do {
+		++iterations;
+		if (read_big_block (fd, buf))
+			goto quit;
+		getitimer(ITIMER_REAL, &e2);
+		elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
+		 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
+	} while (elapsed < 3.0 && iterations < max_iterations);
 
-	elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
-	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
-	if (timing_MB >= elapsed)  /* more than 1MB/s */
-		printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
-			timing_MB, elapsed, timing_MB / elapsed);
+	total_MB = iterations * TIMING_BUF_MB;
+	if ((total_MB / elapsed) > 1.0)  /* more than 1MB/s */
+		printf("%3u MB in %5.2f seconds = %6.2f MB/sec\n",
+			total_MB, elapsed, total_MB / elapsed);
 	else
-		printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
-			timing_MB, elapsed, timing_MB / elapsed * 1024);
-
-	if (elapsed <= (correction * 2))
-		printf("Hmm.. suspicious results: probably not enough free memory for a proper test.\n");
-#if 0  /* the "estimate" is just plain wrong for many systems.. */
-	else if (correction != 0.0) {
-		printf(" Estimating raw driver speed: ");
-		elapsed -= correction;
-		if (timing_MB >= elapsed)  /* more than 1MB/s */
-			printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
-				timing_MB, elapsed, timing_MB / elapsed);
-		else
-			printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
-				timing_MB, elapsed, timing_MB / elapsed * 1024);
-	}
-#endif
+		printf("%3u MB in %5.2f seconds = %6.2f kB/sec\n",
+			total_MB, elapsed, total_MB / elapsed * 1024);
 quit:
 	if (-1 == shmdt(buf))
 		perror ("could not detach sharedmem buf");
@@ -576,7 +575,7 @@ static void interpret_standby (unsigned int standby)
 				if (hrs && mins)  printf(" + ");
 				if (mins)	  printf("%u minutes", mins);
 			} else
-				printf("illegal value)\n");
+				printf("illegal value");
 			break;
 	}
 	printf(")\n");
@@ -698,6 +697,10 @@ static void interpret_xfermode (unsigned int xfermode)
 #define VXVM_MAJOR 199
 #endif
 
+#ifndef CCISS_MAJOR
+#define CCISS_MAJOR 104
+#endif
+
 void process_dev (char *devname)
 {
 	int fd;
@@ -728,13 +731,15 @@ void process_dev (char *devname)
 	case (MD_MAJOR):
 #endif
 	case (VXVM_MAJOR):
+#ifdef LVM_BLK_MAJOR
 	case (LVM_BLK_MAJOR):
+#endif
+	case (CCISS_MAJOR):
 		is_scsi_hd = 1;
 		break;
 #ifdef XT_DISK_MAJOR
 	case (XT_DISK_MAJOR):
 		is_xt_hd = 1;
-		timing_MB = 2;
 		break;
 #endif
 	case IDE0_MAJOR:
@@ -1297,7 +1302,7 @@ void process_dev (char *devname)
 		    __le16_to_cpus((__u16 *)(&args[i]));
 		}    
 		identify((void *)&args[4], NULL);
-identify_abort:
+identify_abort:	;
 	}
 #endif
 #ifdef HDIO_SET_BUSSTATE
@@ -1348,16 +1353,6 @@ identify_abort:
 	}
  #endif /* HDIO_GET_ACOUSTIC */
 
-	//
-	// Try to adjust buffersize based on device size
-	//
-	if (do_ctimings || do_timings) {
-		if (ioctl(fd, BLKGETSIZE, &parm))
-			perror(" BLKGETSIZE failed");
-		else if ( (parm * 512/(1024 * 1024)) < timing_MB)
-			timing_MB = parm * 512/(1024 * 1024);
-	}
-
 	if (do_ctimings)
 		time_cache (fd);
 	if (do_timings)
@@ -1493,9 +1488,9 @@ static int fromhex (unsigned char c)
 
 static int identify_from_stdin (void)
 {
-	unsigned short buf[800];
-	unsigned char *b = (unsigned char *)buf;
-	int i, count = read(0, b, 1280);
+	unsigned short sbuf[800];
+	unsigned char  buf[1600], *b = (unsigned char *)buf;
+	int i, count = read(0, buf, 1280);
 
 	if (count != 1280) {
 		fprintf(stderr, "read(1280 bytes) failed (rc=%d)", count);
@@ -1503,12 +1498,12 @@ static int identify_from_stdin (void)
 		exit(errno);
 	}
 	for (i = 0; count >= 4; ++i) {
-		buf[i] = (fromhex(*b++) << 12) | (fromhex(*b++) << 8) | (fromhex(*b++) << 4) | fromhex(*b++);
-		__le16_to_cpus((__u16 *)(&buf[i]));
-		++b;
+		sbuf[i] = (fromhex(b[0]) << 12) | (fromhex(b[1]) << 8) | (fromhex(b[2]) << 4) | fromhex(b[3]);
+		__le16_to_cpus((__u16 *)(&sbuf[i]));
+		b += 5;
 		count -= 5;
 	}
-	identify((void *)buf, NULL);
+	identify(sbuf, NULL);
 	return 0;
 }
 
