@@ -3,7 +3,13 @@
 #include <errno.h>
 #include <string.h>
 #include <linux/types.h>
+#include <endian.h>
 
+#if __BYTE_ORDER == __BIG_ENDIAN
+#define __USE_XOPEN
+#endif
+
+#include <unistd.h>
 #include "hdparm.h"
 
 /* device types */
@@ -414,11 +420,15 @@ void identify (__u16 *id_supplied, const char *devname)
 	__u16 dev = NO_DEV, eqpt = NO_DEV;
 	__u8  have_mode = 0, err_dma = 0;
 	__u8  chksum = 0;
-	__u32 ll, mm;
+	__u32 ll, mm, nn, oo;
 	__u64 bbbig; /* (:) */
 
 	if (id_supplied) {
+#if __BYTE_ORDER == __BIG_ENDIAN
+		swab(id_supplied, val, sizeof(val));
+#else
 		memcpy(val, id_supplied, sizeof(val));
+#endif
 	} else {
 		id_file = calloc(1, strlen(devname)+1+strlen(fmt));
 		sprintf(id_file, fmt, devname);
@@ -511,7 +521,7 @@ void identify (__u16 *id_supplied, const char *devname)
 	 * defined until ATA-3 & the CDROM std uses different words.) */
 	printf("Standards:");
 	if(eqpt != CDROM) {
-		if(val[MINOR] && (val[MINOR] <=MINOR_MAX)){
+		if(val[MINOR] && (val[MINOR] <= MINOR_MAX)){
 			if(like_std < 3) like_std = 3;
 			std = actual_ver[val[MINOR]];
 			if(std) {
@@ -645,6 +655,12 @@ void identify (__u16 *id_supplied, const char *devname)
 			}
 			if(jj) {
 				mm = (__u32)val[CAPACITY_MSB] << 16 | val[CAPACITY_LSB];
+				if(like_std < 3) { /* check Endian of capacity bytes */
+					nn = val[LCYLS_CUR] * val[LHEADS_CUR] * val[LSECTS_CUR];
+					oo = (__u32)val[CAPACITY_LSB] << 16 | val[CAPACITY_MSB];
+					if(abs(mm - nn) > abs(oo - nn))
+						mm = oo;
+				}
 				printf("\tCHS current addressable sectors:%11u\n",mm);
 			} 
 		}
